import numpy as np


class SingleFFactor:

    def __init__(self, name, distribution, uncertainty):
        """Returns a spectrally independent, self distributed f-factor instance.
        @type name: str
        @type distribution: str
        @type uncertainty: float
        """
        try:
            self._name = name
            self._distribution = distribution[0]
            self.is_valid_distribution()
        except ValueError:
            print('Unknown distribution {0} for {1} -> deleted.'.format(self._distribution, self._name))
            del self
        else:
            self._uncertainty = uncertainty
            self._random_value = 0

    def is_valid_distribution(self):
        return ['n', 'r', 't'].index(self._distribution)

    def gen_random_value(self):
        """Generate a random number from the individual distribution."""
        if self._distribution == 'n':
            return np.random.normal(0, 1)
        elif self._distribution == 'r':
            return np.random.uniform(-1, 1)
        elif self._distribution == 't':
            return np.random.triangular(-1, 0, 1)

    def calc_f(self, random_value: float = None):
        """Generate a random number for the f-factor."""
        self._random_value = random_value if random_value else self.gen_random_value()
        return 1 + self._uncertainty * self._random_value

    def get_random_value(self): return self._random_value


class MultipleFFactor:

    def __init__(self, name, distribution, uncertainty):
        """Returns a spectrally dependent, self distributed f-factor instance.
        @type name: str
        @type distribution: str
        @type uncertainty: np.ndarray[float64]
        """
        try:
            self._name = name
            self._distribution = distribution[0]
            self.is_valid_distribution()
        except ValueError:
            print('Unknown distribution {0} for {1} -> deleted.'.format(self._distribution, self._name))
            del self
        else:
            self._uncertainty = uncertainty
            self._random_value: float = 0

    def is_valid_distribution(self):
        return ['n', 'r', 't'].index(self._distribution)

    def gen_random_value(self):
        """Generate a random number from the individual distribution."""
        if self._distribution == 'n':
            return np.random.normal(0, 1)
        elif self._distribution == 'r':
            return np.random.uniform(-1, 1)
        elif self._distribution == 't':
            return np.random.triangular(-1, 0, 1)

    def calc_f(self, index=None, random_value=None):
        """Returns random values at an optional given index for the f-factor. If random_value is none, a new random value is generated.
        @type index: int
        @type random_value: float
        """
        self._random_value = random_value if random_value else self.gen_random_value()

        # index is None -> spectrally correlated f-factor calculated in the outer loop
        if index is None:
            return 1 + self._uncertainty * self._random_value
        # index is not None -> not correlated f-factor calculated in the inner loop (f_noise)
        else:
            return 1 + self._uncertainty[index] * self._random_value

    def get_random_value(self): return self._random_value
