"""General informations"""

# Code example
# Publication: C. Schinke, H. Pollex, D. Hinken, M. Wolf, K. Bothe, I. Kröger, S. Nevas and S. Winter,
# Calibrating Spectrometers for Measurements of the Spectral Irradiance caused by Solar Radiation,
# Metrologia (2020), https://doi.org/10.1088/1681-7575/abafc5
# License for this code: CC BY 4.0 (see LICENSE.txt)



# The parameters for the uncertainty analysis can be adapted by changing the assignments in lines 34 to 79
# or by changing the content of the files located in the "DataFiles" directory.

# System requirements: Python 3.6 or above with packages numpy, scipy and mathplotlib.
# Refer to the Python documentation for installing these packages in your Python distribution.


"""Import Packages"""

from f_factor_class import SingleFFactor, MultipleFFactor
import os
import math
import time
import numpy as np
import scipy.constants as const
import matplotlib.pyplot as plt

"""Define parameters for calculation"""

# Path for data files
Data_File_Directory = os.path.abspath('DataFiles') + '/'

# Monte-Carlo iterations
n_MC = 1000

# Define detector for calculation (comment/uncomment lines)
detector = 'Si'
# detector = 'InGaAs'

# Define wavelength range for output data (start/stop in nm)
WL_range = [300, 1650]

"""Define parameters of the measurement data"""

# Determine number of measurement averaging
averages = 25

# Set integration times [ms] used for the measurement
tint_ref = {'Si': 1150, 'InGaAs': 680}
tint_DUT = {'Si': 50, 'InGaAs': 125}

# Set the measuring distance in the reference measurement [m]
meas_dist_ref = 0.3

print('Finished parameter definition.')

"""Import uncertainty data"""

# Set scalar uncertainty calculation parameters (Table 1)
t_meas_total = 50  # [s]
u_darkdrift = {'Si': 9.865e-6 * t_meas_total, 'InGaAs': 5.733e-7 * t_meas_total}
u_rep = {'Si': 4.347e-3, 'InGaAs': 1.951e-3}
u_lintint = {'Si': 6.785e-4, 'InGaAs': 1.809e-4}
u_linirr = {'Si': 4.536e-4, 'InGaAs': 3.163e-3}
dlambda = {'Si': 0.15, 'InGaAs': 0.15}
deltalambda = {'Si': 3.5, 'InGaAs': 7.0}
u_cur = {'Si': 7e-4, 'InGaAs': 7e-4}
u_tilt = {'Si': 0.6, 'InGaAs': 0.6}
u_temp = {'Si': 0.3, 'InGaAs': 0}
u_dist = 0.00088

# Import spectral uncertainty calculation parameters (Figure 10)
u_strayext = np.loadtxt(fname=Data_File_Directory + 'u_strayext.txt', dtype=float, delimiter='\t', skiprows=1,
                        usecols=1)
cT_lambda = np.loadtxt(fname=Data_File_Directory + 'cT_lambda.txt', dtype=float, delimiter='\t', skiprows=1, usecols=1)
u_strayint_DUT = np.loadtxt(fname=Data_File_Directory + 'u_strayint_{0}.txt'.format(detector), dtype=float,
                            delimiter='\t', skiprows=1, usecols=2)
u_strayint_ref = np.loadtxt(fname=Data_File_Directory + 'u_strayint_{0}.txt'.format(detector), dtype=float,
                            delimiter='\t', skiprows=1, usecols=1)

print('Finished uncertainty data import.')


"""Import Measurement and reference data"""

# Import and determine DUT measurement data
DUT_meas_data = np.loadtxt(fname=Data_File_Directory + 'DUT_measurement_signal_{0}.txt'.format(detector), dtype=float,
                           delimiter='\t', skiprows=1)
WL = DUT_meas_data[:, 0]
S_2_DUT = (DUT_meas_data[:, 1] - DUT_meas_data[:, 3]) / tint_DUT[detector]
u_noise_DUT = np.sqrt(DUT_meas_data[:, 2] ** 2 + DUT_meas_data[:, 4] ** 2) / np.sqrt(averages) / tint_DUT[detector]

# Import and determine Ref measurement data
ref_meas_data = np.loadtxt(fname=Data_File_Directory + 'Ref_measurement_signal_{0}.txt'.format(detector), dtype=float,
                           delimiter='\t', skiprows=1)
S_2_ref = (ref_meas_data[:, 1] - ref_meas_data[:, 3]) / tint_ref[detector]
u_noise_ref = np.sqrt(ref_meas_data[:, 2] ** 2 + ref_meas_data[:, 4] ** 2) / np.sqrt(averages) / tint_ref[detector]

# Import and determine reference calibration data
ref_cal_data = np.loadtxt(fname=Data_File_Directory + 'Ref_spectral_irradiance.txt', dtype=float,
                          delimiter='\t', skiprows=1)
WL_cal = ref_cal_data[:, 0]
E_ref = ref_cal_data[:, 1]
u_Eref = ref_cal_data[:, 2]

# Import Spectral Stray Light Correction Matrices
SSLC_Matrix = np.loadtxt(fname=Data_File_Directory + 'SpectralStrayLightCorrectionMatrix_{0}.txt'.format(detector),
                         dtype=float, delimiter='\t')

print('Finished measurement data import.')


"""Measurement signal calculation and interpolation of data"""


# Define helper functions

def spectral_stray_light_correction(S):
    """Returns the spectral stray light corrected signal S."""
    return np.matmul(SSLC_Matrix, S)


def spectral_bandwidth_correction(S):
    """Returns the spectral bandwidth corrected signal S."""
    return S - ((-2 * S + np.interp(WL - deltalambda[detector]/2, WL, S) + np.interp(WL + deltalambda[detector]/2, WL, S)) / 12)


def cut_data(array):
    """Returns the array data inside the given range of the calibration data (typically 250nm-1700nm)."""
    return array[(WL >= WL_cal[0]) & (WL <= WL_cal[-1])]


# calculate signals S_X

# No nonlinearity correction required (Equations 5, 11)
S_3_DUT = S_2_DUT
S_3_ref = S_2_ref

# Apply spectral stray light correction to S_DUT and S_ref (Equations 6, 11)
S_4_DUT = spectral_stray_light_correction(S_3_DUT)
S_4_ref = spectral_stray_light_correction(S_3_ref)

# Apply spectral bandwidth correction to S_DUT and S_ref (Equations 7, 11)
S_5_DUT = spectral_bandwidth_correction(S_4_DUT)
S_5_ref = spectral_bandwidth_correction(S_4_ref)


# Cut all data that is outside the given wavelengths of the reference calibration data
S_2_DUT = cut_data(S_2_DUT)
S_2_ref = cut_data(S_2_ref)
S_4_DUT = cut_data(S_4_DUT)
S_4_ref = cut_data(S_4_ref)
S_5_DUT = cut_data(S_5_DUT)
S_5_ref = cut_data(S_5_ref)
u_noise_DUT = cut_data(u_noise_DUT)
u_noise_ref = cut_data(u_noise_ref)
u_strayint_DUT = cut_data(u_strayint_DUT)
u_strayint_ref = cut_data(u_strayint_ref)
WL = cut_data(WL)

# Interpolate spectral uncertainties and reference calibration data to WL
E_ref = np.interp(WL, WL_cal, E_ref)
u_Eref = np.interp(WL, WL_cal, u_Eref)
u_strayext = np.interp(WL, WL_cal, u_strayext)
cT_lambda = np.interp(WL, WL_cal[:-6], cT_lambda)

# Calculate the radiometric correction function K_lambda (Equation 10)
K_lambda = E_ref / S_5_ref

# Calculate spectral irradiance E_DUT (Equation 12)
E_DUT = S_5_DUT * K_lambda

print('Finished signal calculation and interpolation of uncertainties and reference data.')


"""Uncertainty analysis with Monte-Carlo-Simulation"""


def E_bb(wl, T):                                                                  # Equation 37
    """Returns the spectral irradiance of a black body at given temperature T in [K] and wavelength wl in [m].
    @type wl: np.ndarray[float64]
    @type T: float
    """
    return (2 * const.pi * const.h * const.c) / (wl ** 5 * (np.expm1(const.h * const.c / (wl * const.k * T)) - 1))


def get_f_cur(delta_I, wl):
    """Returns a value for f_cur.
    @type delta_I: float
    @type wl: np.ndarray[float64]
    """
    wl = wl * 1e-9                                                                 # nm -> m
    V_op = 19.42813                                                                # V
    I_op = 9.7                                                                     # A
    R_0 = 0.148                                                                    # Ohm
    T_0 = 296.03                                                                   # K
    alpha = 4.4e-3                                                                 # 1/K
    T_op = (V_op + I_op * R_0 * (T_0 * alpha - 1)) / (I_op * R_0 * alpha)          # Equation 41
    delta_T = (V_op / (I_op ** 2 * R_0 * alpha)) * delta_I                         # Equation 42
    return 1 + (E_bb(wl, T_op - delta_T) - E_bb(wl, T_op)) / E_bb(wl, T_op)        # Equation 45


# Initialize resulting Matrix (n: iteration, k: wavelength)
E_MC_DUT = np.zeros(shape=(n_MC, len(WL)))  # -> E_MC_DUT[n,k]

# Initialize spectral (single) and non-spectral (multiple) f-factors with the associated uncertainties
# pass 'n', 'r' or 't' for normal, rectangular or triangular distribution
f_noise_DUT = MultipleFFactor('f_noise_DUT', 'n', u_noise_DUT / S_2_DUT)
f_noise_ref = MultipleFFactor('f_noise_ref', 'n', u_noise_ref / S_2_ref)
f_dark_DUT = MultipleFFactor('f_dark_DUT', 'r', u_darkdrift[detector] / S_2_DUT)
f_dark_ref = MultipleFFactor('f_dark_ref', 'r', u_darkdrift[detector] / S_2_ref)
f_strayint_DUT = MultipleFFactor('f_strayint_DUT', 'r', u_strayint_DUT)
f_strayint_ref = MultipleFFactor('f_strayint_ref', 'r', u_strayint_ref)
f_strayext_ref = MultipleFFactor('f_strayext_ref', 'r', u_strayext)
f_lintint_DUT = SingleFFactor('f_lintint_DUT', 'r', u_lintint[detector])
f_lintint_ref = SingleFFactor('f_lintint_ref', 'r', u_lintint[detector])
f_linirr_DUT = SingleFFactor('f_linirr_DUT', 'r', u_linirr[detector])
f_linirr_ref = SingleFFactor('f_linirr_ref', 'r', u_linirr[detector])
f_bw_DUT = MultipleFFactor('f_bw_DUT', 't', (-2 * S_5_DUT + np.interp(WL - deltalambda[detector] / 2, WL, S_5_DUT) + np.interp(WL + deltalambda[detector] / 2, WL, S_5_DUT)) / 12 / S_5_DUT)
f_bw_ref = MultipleFFactor('f_bw_ref', 't', (-2 * S_5_ref + np.interp(WL - deltalambda[detector] / 2, WL, S_5_ref) + np.interp(WL + deltalambda[detector] / 2, WL, S_5_ref)) / 12 / S_5_ref)
f_wl_DUT = MultipleFFactor('f_wl_DUT', 'r', np.gradient(S_4_DUT, WL) * dlambda[detector] / S_4_DUT)
f_wl_ref = MultipleFFactor('f_wl_ref', 'r', np.gradient(S_4_ref, WL) * dlambda[detector] / S_4_ref)
f_dist = SingleFFactor('f_dist', 'r', -2 * u_dist / meas_dist_ref)
f_rep = SingleFFactor('f_rep', 'r', u_rep[detector])
f_cur = SingleFFactor('f_cur', 'r', u_cur[detector])
f_temp = MultipleFFactor('f_temp', 'r', u_temp[detector] * cT_lambda)
f_Eref = MultipleFFactor('f_ref', 'n', u_Eref / E_ref)
f_tilt_DUT = SingleFFactor('f_tilt_DUT', 'r', math.radians(u_tilt[detector]) ** 2 / 2)
f_tilt_ref = SingleFFactor('f_tilt_ref', 'r', math.radians(u_tilt[detector]) ** 2 / 2)

# Loop through Monte-Carlo iterations
print('Run Monte-Carlo analysis ({0} iterations)...'.format(n_MC))
for n in range(n_MC):

    # Calculate random f-factor values/vectors
    f_dark_DUT_n = f_dark_DUT.calc_f()
    f_dark_ref_n = f_dark_ref.calc_f()
    f_strayint_DUT_n = f_strayint_DUT.calc_f()
    f_strayint_ref_n = f_strayint_ref.calc_f()
    f_strayext_ref_n = f_strayext_ref.calc_f()
    f_lintint_DUT_n = f_lintint_DUT.calc_f()
    f_lintint_ref_n = f_lintint_ref.calc_f(random_value=f_lintint_DUT.get_random_value())  # -> correlation between DUT and ref
    f_linirr_DUT_n = f_linirr_DUT.calc_f()
    f_linirr_ref_n = f_linirr_ref.calc_f(random_value=f_linirr_DUT.get_random_value())  # -> correlation between DUT and ref
    f_bw_DUT_n = f_bw_DUT.calc_f()
    f_bw_ref_n = f_bw_ref.calc_f(random_value=f_bw_DUT.get_random_value())  # -> correlation between DUT and ref
    f_wl_DUT_n = f_wl_DUT.calc_f()
    f_wl_ref_n = f_wl_ref.calc_f(random_value=f_wl_DUT.get_random_value())  # -> correlation between DUT and ref
    f_dist_n = f_dist.calc_f()
    f_rep_n = f_rep.calc_f()
    f_cur_n = get_f_cur(delta_I=f_cur.gen_random_value() * u_cur[detector], wl=WL)
    f_temp_n = f_temp.calc_f()
    f_Eref_n = f_Eref.calc_f()
    f_tilt_DUT_n = f_tilt_DUT.calc_f()
    f_tilt_ref_n = f_tilt_ref.calc_f()

    # Loop through wavelengths
    for k in range(len(WL)):
        f_noise_DUT_n_k = f_noise_DUT.calc_f(index=k)
        f_noise_ref_n_k = f_noise_ref.calc_f(index=k)

        # Summarized f-factor (Equation 13)
        f_MC = ((f_noise_DUT_n_k * f_dark_DUT_n[k] * f_lintint_DUT_n * f_linirr_DUT_n * f_strayint_DUT_n[k] *
                 f_bw_DUT_n[k] * f_wl_DUT_n[k] * f_tilt_DUT_n) /
                (f_noise_ref_n_k * f_dark_ref_n[k] * f_lintint_ref_n * f_linirr_ref_n * f_strayint_ref_n[k] *
                 f_strayext_ref_n[k] * f_bw_ref_n[k] * f_wl_ref_n[k] * f_tilt_ref_n)) * f_dist_n * f_rep_n * f_Eref_n[k] * f_cur_n[k] * f_temp_n[k]

        # Calculate the DUT spectral irradiance for this iteration and wavelength (Equation 14)
        E_MC_DUT[n, k] = E_DUT[k] * f_MC


# Monte-Carlo calculation finished
# Calculate the standard deviation at each wavelength (axis = 0) and derive uncertainty data (Equation 16)
u_E_DUT = np.std(E_MC_DUT, axis=0) / E_DUT * 200   # -> rel. uncertainty in % for k=2

# Cut data outside defined wavelength range
E_DUT = E_DUT[(WL >= WL_range[0]) & (WL <= WL_range[1])]
u_E_DUT = u_E_DUT[(WL >= WL_range[0]) & (WL <= WL_range[1])]
WL = WL[(WL >= WL_range[0]) & (WL <= WL_range[1])]

print('Finished Monte-Carlo-Simulation and resulting uncertainty calculation.')

"""Data export"""

# Define the file header
File_Header = 'wavelength [nm]\tspectral irradiance[W/m^2/nm]\trel. uncertainty [k=2] [%]'

# Determine the output data
output_data = np.array([WL, E_DUT, u_E_DUT]).transpose()

# Export the data into a txt-File
np.savetxt(fname='mc-result_{0}_{1}.txt'.format(detector, time.strftime('%d%m%Y')), X=output_data, fmt=['%.2f', '%.9f', '%.6f'], delimiter='\t', header=File_Header, comments='')

print('Finished data export.')

"""Plot the results"""

# Create a figure
fig, ax1 = plt.subplots(num='DEVICE UNDER TEST', figsize=(10, 8), dpi=80, facecolor='w', edgecolor='k')
fig.suptitle('Spectral irradiance of DUT', fontsize=24, fontweight='bold')

# Plot the spectral irradiance data of the DUT on a first y-axis
irr_color = 'tab:blue'
ax1.plot(WL, E_DUT, color=irr_color, linewidth=1.5, alpha=0.6)
ax1.set_xlabel("wavelength " + r"$\lambda$" + " [nm]", fontsize=12)
ax1.set_xlim(round(WL.min(), 0), round(WL.max(), 0))
ax1.tick_params(axis='x', labelsize=12)
ax1.set_ylabel("spectral irradiance " + r"$E_{DUT} \ [W/m^2/nm]$", color=irr_color, fontsize=12)
ax1.set_ylim(0, 2)
ax1.tick_params(axis='y', labelcolor=irr_color, labelsize=12)

# Plot the relative uncertainty [k=2] of the DUT's spectral irradiance data on a second y-axis
u_color = 'tab:red'
ax2 = ax1.twinx()
ax2.plot(WL, u_E_DUT, color='r', linewidth=1, linestyle='dashed')
ax2.set_ylabel("relative uncertainty " + r"$U \ (E_{DUT})$" + " [k=2] [%]", color=u_color, fontsize=12)
ax2.set_ylim(0, 20)
ax2.tick_params(axis='y', labelcolor=u_color, labelsize=12)

plt.grid(b=True, which='major', axis='y')
plt.show()

print('Finished plotting.')
print('Finished script.')
